package com.autoussd.autoussdflutter

import android.util.Log
import androidx.activity.ComponentActivity
import androidx.annotation.NonNull
import com.autoussd.AutoUssd
import com.autoussd.models.Network

import io.flutter.embedding.engine.plugins.FlutterPlugin
import io.flutter.embedding.engine.plugins.activity.ActivityAware
import io.flutter.embedding.engine.plugins.activity.ActivityPluginBinding
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.MethodChannel.MethodCallHandler
import io.flutter.plugin.common.MethodChannel.Result as MethodChannelResult

class AutoUssdFlutterPlugin : FlutterPlugin, MethodCallHandler, ActivityAware {
  companion object {
    private const val TAG = "AutoUssdFlutterPlugin"
    private const val NATIVE_COUNT_CALLBACK_NAME = "autoussdCount"
    private const val NATIVE_RESULT_CALLBACK_NAME = "autoussdResult"
    private const val NATIVE_EXECUTE_SESSION_METHOD = "executeSession"
    private const val NATIVE_GET_DEVICE_SIM_NETWORKS_METHOD = "getDeviceSimNetworks"
    private const val FLUTTER_COUNT_CALLBACK_METHOD = "onCount"
    private const val FLUTTER_RESULT_CALLBACK_METHOD = "onResult"
  }

  private lateinit var channel: MethodChannel

  override fun onAttachedToEngine(@NonNull flutterPluginBinding: FlutterPlugin.FlutterPluginBinding) {
    Log.d(TAG, "On Attached To Engine")

    channel = MethodChannel(
      flutterPluginBinding.binaryMessenger,
      "autoussdflutter"
    )

    channel.setMethodCallHandler(this)
  }

  override fun onAttachedToActivity(binding: ActivityPluginBinding) {
    Log.d(TAG, "On Attached To Activity")

    try {
      AutoUssd.init(binding.activity as ComponentActivity)

      AutoUssd.onSessionCount(NATIVE_COUNT_CALLBACK_NAME) {
        binding.activity.runOnUiThread {
          channel.invokeMethod(
            FLUTTER_COUNT_CALLBACK_METHOD,
            it
          )
        }
      }

      AutoUssd.onSessionResult(NATIVE_RESULT_CALLBACK_NAME) {
        binding.activity.runOnUiThread {
          channel.invokeMethod(
            FLUTTER_RESULT_CALLBACK_METHOD,
            listOf(
              it.sessionId,
              it.status.toString(),
              it.description,
              it.lastContent,
              it.parser,
            )
          )
        }
      }

      Log.d(TAG, "SDK initialized")
    } catch (err: Exception) {
      err.printStackTrace()
      Log.e(TAG, err.message ?: "Could not create instance of AutoUssd")
    }
  }

  override fun onDetachedFromActivity() {
    Log.d(TAG, "On Detached From Activity")

    AutoUssd.offSessionCount(NATIVE_COUNT_CALLBACK_NAME)
    AutoUssd.offSessionResult(NATIVE_RESULT_CALLBACK_NAME)
    AutoUssd.dispose()
  }

  override fun onReattachedToActivityForConfigChanges(binding: ActivityPluginBinding) {
    Log.d(TAG, "On Reattached To Activity For Config Changes")

    channel.setMethodCallHandler(this)
  }

  override fun onDetachedFromActivityForConfigChanges() {
    Log.d(TAG, "On Detached To Activity For Config Changes")

    channel.setMethodCallHandler(null)
  }

  override fun onDetachedFromEngine(@NonNull binding: FlutterPlugin.FlutterPluginBinding) {
    Log.d(TAG, "On Detached From Engine")

    channel.setMethodCallHandler(null)
  }

  override fun onMethodCall(@NonNull call: MethodCall, @NonNull result: MethodChannelResult) {
    Log.d(TAG, "On Method Call")

    if (call.method == NATIVE_EXECUTE_SESSION_METHOD) {
      val sessionId = call.argument<String?>("sessionId")!!
      val variables = call.argument<List<String>?>("variables")?.toTypedArray()
      val network = call.argument<HashMap<String, String>?>("network")

      if (network == null) {
        AutoUssd.executeSession(
          sessionId,
          variables,
          null,
        )
      } else {
        AutoUssd.executeSession(
          sessionId,
          variables,
          Network(
            network["mcc"],
            network["mnc"],
          ),
        )
      }
    } else if (call.method == NATIVE_GET_DEVICE_SIM_NETWORKS_METHOD) {
      // The host app must ensure the READ_PHONE_STATE
      // permission is granted prior to calling this method
      val networks = AutoUssd.getDeviceSimNetworks()

      result.success(
        networks.map {
          mapOf(
            "name" to it.name,
            "mcc" to it.mcc,
            "mnc" to it.mnc,
            "countryName" to it.countryName,
            "countryIsoCode" to it.countryIsoCode,
            "countryDialCode" to it.countryDialCode,
          )
        }
      )
    } else result.notImplemented()
  }
}
