import 'dart:async';

import 'package:flutter/services.dart';

typedef SessionCountCallback = void Function(int count);

typedef SessionResultCallback = void Function(Result result);

class AutoUssdFlutter {
  static const String _countCallbackMethod = "onCount";
  static const String _resultCallbackMethod = "onResult";
  static const String _executeSessionMethod = "executeSession";
  static const String _getDeviceSimNetworksMethod = "getDeviceSimNetworks";
  static const MethodChannel _CHANNEL = const MethodChannel(
    "autoussdflutter",
  );
  static AutoUssdFlutter? _instance;

  /// Throws an [Exception] if [AutoUssdFlutter.init]
  /// has not ben previously called resulting in
  /// [_instance] being null.
  static void _checkInstance() {
    if (_instance == null) {
      throw new Exception(
        "Instance is null. Run AutoUssdFlutter.init() first",
      );
    }
  }

  /// Initializes the static AutoUssdFlutter singleton
  /// This should be called early on in the app runtime
  static void init() {
    if (_instance == null) {
      _instance = AutoUssdFlutter._();
    }
  }

  /// Registers a [SessionCountCallback]
  /// which is called when the session
  /// count in the database changes.
  static void onSessionCount(
    SessionCountCallback callback,
  ) {
    _checkInstance();
    _instance!._onSessionCount(callback);
  }

  /// Removes a [SessionCountCallback]
  static void offSessionCount(
    SessionCountCallback callback,
  ) {
    _checkInstance();
    _instance!._offSessionCount(callback);
  }

  /// Registers a [SessionResultCallback]
  /// which is called when a session execution
  /// result is available.
  static void onSessionResult(
    SessionResultCallback callback,
  ) {
    _checkInstance();
    _instance!._onSessionResult(callback);
  }

  /// Removes a [SessionResultCallback]
  static void offSessionResult(
    SessionResultCallback callback,
  ) {
    _checkInstance();
    _instance!._offSessionResult(callback);
  }

  /// Executes a session matching
  /// the specified in your app.
  static void executeSession(
    String sessionId, [
    List<String>? variables,
    Network? network,
  ]) {
    _checkInstance();
    _instance!._executeSession(
      sessionId,
      variables,
      network,
    );
  }

  /// Retrieves a list of [Network] of SIMs in the device
  /// Requires the READ\_PHONE\_STATE permission to
  /// have already been granted
  static Future<List<Network>> getDeviceSimNetworks() {
    _checkInstance();
    return _instance!._getDeviceSimNetworks();
  }

  final List<SessionCountCallback> _countCallbacks = [];
  final List<SessionResultCallback> _resultCallbacks = [];

  AutoUssdFlutter._() {
    _CHANNEL.setMethodCallHandler((call) async {
      if (call.method == _countCallbackMethod) {
        _notifyCountCallbacks(call);
      } else if (call.method == _resultCallbackMethod) {
        _notifyResultCallbacks(call);
      }
    });
  }

  /// Registers a [SessionCountCallback]
  /// which is called when the session
  /// count in the database changes.
  void _onSessionCount(
    SessionCountCallback callback,
  ) {
    if (!_countCallbacks.contains(callback)) {
      _countCallbacks.add(callback);
    }
  }

  /// Removes a [SessionCountCallback]
  void _offSessionCount(
    SessionCountCallback callback,
  ) {
    if (_countCallbacks.contains(callback)) {
      _countCallbacks.remove(callback);
    }
  }

  /// Registers a [SessionResultCallback]
  /// which is called when a session execution
  /// result is available.
  void _onSessionResult(
    SessionResultCallback callback,
  ) {
    if (!_resultCallbacks.contains(callback)) {
      _resultCallbacks.add(callback);
    }
  }

  /// Removes a [SessionResultCallback]
  void _offSessionResult(
    SessionResultCallback callback,
  ) {
    if (_resultCallbacks.contains(callback)) {
      _resultCallbacks.remove(callback);
    }
  }

  /// Notifies count callbacks of changes in session count
  void _notifyCountCallbacks(MethodCall call) {
    final count = call.arguments as int;

    for (final callback in _countCallbacks) {
      callback(count);
    }
  }

  /// Notifies result callbacks of session execution results
  void _notifyResultCallbacks(MethodCall call) {
    final List<dynamic> args = call.arguments;
    final sessionId = args[0] as String;
    final status = args[1] as String;
    final description = args[2] as String;
    final lastContent = args[3] as String?;
    final parser = args[4] as String?;

    for (final callback in _resultCallbacks) {
      callback(
        Result._(
          sessionId,
          status,
          description,
          lastContent,
          parser,
        ),
      );
    }
  }

  /// Executes a session matching
  /// the specified in your app.
  Future<void> _executeSession(
    String sessionId, [
    List<String>? variables,
    Network? network,
  ]) async {
    final args = {
      "sessionId": sessionId,
      "variables": variables,
    };

    if (network != null) {
      args["network"] = network.toMap();
    }

    await _CHANNEL.invokeMethod(
      _executeSessionMethod,
      args,
    );
  }

  /// Retrieves a list of [Network] of SIMs in the device
  /// Requires the READ\_PHONE\_STATE permission to
  /// have already been granted
  Future<List<Network>> _getDeviceSimNetworks() async {
    final networksListData = await _CHANNEL.invokeMethod(
      _getDeviceSimNetworksMethod,
    ) as List<dynamic>?;

    if (networksListData != null) {
      return networksListData.map((map) {
        return Network(
          map["name"],
          map["mcc"],
          map["mnc"],
          map["countryName"],
          map["countryIsoCode"],
          map["countryDialCode"],
        );
      }).toList(growable: false);
    } else {
      return [];
    }
  }
}

/// Represents the result
/// of a session execution
///
/// After the SDK executes a session,
/// it asynchronously returns an object
/// of this type which contains information
/// on what went on during the execution
class Result {
  late final String sessionId;
  late final ResultStatus status;
  late final String description;
  late final String? lastContent;
  late final String? parser;

  Result._(
    this.sessionId,
    String status,
    this.description,
    this.lastContent,
    this.parser,
  ) {
    this.status = Result._parseStatus(
      status,
    );
  }

  static ResultStatus _parseStatus(String status) {
    switch (status) {
      case "COMPLETED":
        return ResultStatus.COMPLETED;
      case "PARSED":
        return ResultStatus.PARSED;
      case "INVALID_SESSION":
        return ResultStatus.INVALID_SESSION;
      case "UNSUPPORTED_SIM":
        return ResultStatus.UNSUPPORTED_SIM;
      case "SESSION_TIMEOUT":
        return ResultStatus.SESSION_TIMEOUT;
      case "MENU_CONTENT_MISMATCH":
        return ResultStatus.MENU_CONTENT_MISMATCH;
      case "ACCOUNT_SUBSCRIPTION_EXPIRED":
        return ResultStatus.ACCOUNT_SUBSCRIPTION_EXPIRED;
      case "UNKNOWN_ERROR":
        return ResultStatus.UNKNOWN_ERROR;
      default:
        throw "Unknown status: $status";
    }
  }
}

/// Enum constants for report statuses
enum ResultStatus {
  /// Indicates that a session (without SMS parsers)
  /// executed successfully
  COMPLETED,

  /// Indicates that a session execution pending
  /// SMS confirmation has been parsed
  PARSED,

  /// Indicates that the session does
  /// not exist in app's sessions
  INVALID_SESSION,

  /// Indicates that none of the
  /// device's SIM is supported by
  /// the session
  UNSUPPORTED_SIM,

  /// Indicates that the session
  /// timed-out during execution
  SESSION_TIMEOUT,

  /// Indicates that the menu match text didn't
  /// match that of the USSD dialog content
  MENU_CONTENT_MISMATCH,

  /// Indicates that the session owner's
  /// subscription period has elapsed
  ACCOUNT_SUBSCRIPTION_EXPIRED,

  /// Indicates that an unknown error occurred
  UNKNOWN_ERROR,
}

/// Represents a SIM network
class Network {
  final String name;
  final String mcc;
  final String mnc;
  final String countryName;
  final String countryIsoCode;
  final String countryDialCode;

  const Network(
    this.name,
    this.mcc,
    this.mnc,
    this.countryName,
    this.countryIsoCode,
    this.countryDialCode,
  );

  factory Network.fromMap(Map<String, dynamic> map) {
    return Network(
      map["name"] ?? "",
      map["mcc"] ?? "",
      map["mnc"] ?? "",
      map["countryName"] ?? "",
      map["countryIsoCode"] ?? "",
      map["countryDialCode"] ?? "",
    );
  }

  Map<String, String> toMap() {
    return {
      "name": name,
      "mcc": mcc,
      "mnc": mnc,
      "countryName": countryName,
      "countryIsoCode": countryIsoCode,
      "countryDialCode": countryDialCode,
    };
  }
}
